@@ -23,15 +23,25 @@ from googleapiclient import discovery
 from googleapiclient import errors
 
 from . import gcp
+from . import machine_config
+
+try:
+    from tensorflow.python.framework.versions import VERSION
+except ImportError:
+    # Use TF runtime version 2.1 (latest supported) as the default.
+    # https://cloud.google.com/ai-platform/training/docs/runtime-version-list#tpu-support
+    VERSION = "2.1"
 
 
-def deploy_job(region,
-               image_uri,
-               chief_config,
-               worker_count,
-               worker_config,
-               entry_point_args,
-               enable_stream_logs):
+def deploy_job(
+    region,
+    image_uri,
+    chief_config,
+    worker_count,
+    worker_config,
+    entry_point_args,
+    enable_stream_logs,
+):
     """Deploys job with the given parameters to Google Cloud.
 
     Args:
@@ -54,32 +64,41 @@ def deploy_job(region,
     """
     job_id = _generate_job_id()
     project_id = gcp.get_project_name()
-    ml_apis = discovery.build('ml', 'v1', cache_discovery=False)
+    ml_apis = discovery.build("ml", "v1", cache_discovery=False)
 
     request_dict = _create_request_dict(
-        job_id, region, image_uri, chief_config, worker_count, worker_config,
-        entry_point_args)
+        job_id,
+        region,
+        image_uri,
+        chief_config,
+        worker_count,
+        worker_config,
+        entry_point_args,
+    )
     try:
-        response = ml_apis.projects().jobs().create(
-            parent='projects/{}'.format(project_id),
-            body=request_dict
-        ).execute()
+        response = (
+            ml_apis.projects()
+            .jobs()
+            .create(parent="projects/{}".format(project_id), body=request_dict)
+            .execute()
+        )
         _print_logs_info(job_id, project_id)
         if enable_stream_logs:
             _stream_logs(job_id)
     except errors.HttpError as err:
-        raise RuntimeError(
-            'There was an error submitting the job.' + err._get_reason())
+        raise RuntimeError("There was an error submitting the job." + err._get_reason())
     return job_id
 
 
-def _create_request_dict(job_id,
-                         region,
-                         image_uri,
-                         chief_config,
-                         worker_count,
-                         worker_config,
-                         entry_point_args):
+def _create_request_dict(
+    job_id,
+    region,
+    image_uri,
+    chief_config,
+    worker_count,
+    worker_config,
+    entry_point_args,
+):
     """Creates request dictionary for the CAIP training service.
 
     Args:
@@ -100,45 +119,59 @@ def _create_request_dict(job_id,
         The job request dictionary.
     """
     training_input = {}
-    training_input['region'] = region
-    training_input['scaleTier'] = 'custom'
-    training_input['masterType'] = gcp.get_machine_type(
-        chief_config.cpu_cores, chief_config.memory)
+    training_input["region"] = region
+    training_input["scaleTier"] = "custom"
+    training_input["masterType"] = gcp.get_machine_type(
+        chief_config.cpu_cores, chief_config.memory, chief_config.accelerator_type
+    )
 
     # Set master config
     chief_machine_config = {}
-    chief_machine_config['imageUri'] = image_uri
-    chief_machine_config['acceleratorConfig'] = {}
-    chief_machine_config['acceleratorConfig']['count'] = str(
-        chief_config.accelerator_count)
-    chief_machine_config['acceleratorConfig']['type'] = (
-        gcp.get_accelerator_type(chief_config.accelerator_type.value))
+    chief_machine_config["imageUri"] = image_uri
+    chief_machine_config["acceleratorConfig"] = {}
+    chief_machine_config["acceleratorConfig"]["count"] = str(
+        chief_config.accelerator_count
+    )
+    chief_machine_config["acceleratorConfig"]["type"] = gcp.get_accelerator_type(
+        chief_config.accelerator_type.value
+    )
 
-    training_input['masterConfig'] = chief_machine_config
-    training_input['workerCount'] = str(worker_count)
+    training_input["masterConfig"] = chief_machine_config
+    training_input["workerCount"] = str(worker_count)
 
     if worker_count > 0:
-        training_input['workerType'] = gcp.get_machine_type(
-            worker_config.cpu_cores, worker_config.memory)
+        training_input["workerType"] = gcp.get_machine_type(
+            worker_config.cpu_cores,
+            worker_config.memory,
+            worker_config.accelerator_type,
+        )
 
         worker_machine_config = {}
-        worker_machine_config['imageUri'] = image_uri
-        worker_machine_config['acceleratorConfig'] = {}
-        worker_machine_config['acceleratorConfig']['count'] = str(
-            worker_config.accelerator_count)
-        worker_machine_config['acceleratorConfig']['type'] = (
-            gcp.get_accelerator_type(worker_config.accelerator_type.value))
-        training_input['workerConfig'] = worker_machine_config
+        worker_machine_config["imageUri"] = image_uri
+        worker_machine_config["acceleratorConfig"] = {}
+        worker_machine_config["acceleratorConfig"]["count"] = str(
+            worker_config.accelerator_count
+        )
+        worker_machine_config["acceleratorConfig"]["type"] = gcp.get_accelerator_type(
+            worker_config.accelerator_type.value
+        )
+
+        if machine_config.is_tpu_config(worker_config):
+            # AI Platform runtime version spec is required for training
+            # on cloud TPUs.
+            v = VERSION.split(".")
+            worker_machine_config["tpuTfVersion"] = v[0] + "." + v[1]
+        training_input["workerConfig"] = worker_machine_config
 
     if entry_point_args is not None:
-        training_input['args'] = entry_point_args
+        training_input["args"] = entry_point_args
 
     # This is temporarily required so that the `TF_CONFIG` generated by
     # CAIP uses the keyword 'chief' instead of 'master'.
-    training_input['use_chief_in_tf_config'] = True
+    training_input["use_chief_in_tf_config"] = True
     request_dict = {}
-    request_dict['jobId'] = job_id
-    request_dict['trainingInput'] = training_input
+    request_dict["jobId"] = job_id
+    request_dict["trainingInput"] = training_input
     return request_dict
 
 
@@ -149,11 +182,14 @@ def _print_logs_info(job_id, project_id):
         job_id: The job id to print.
         project_id: The project id that is used to generate job access url.
     """
-    print('Job submitted successfully.')
-    print('Your job ID is: ', job_id)
-    print('Please access your job logs at the following URL:')
-    print('https://console.cloud.google.com/mlengine/jobs/{}?project={}'
-          .format(job_id, project_id))
+    print("Job submitted successfully.")
+    print("Your job ID is: ", job_id)
+    print("Please access your job logs at the following URL:")
+    print(
+        "https://console.cloud.google.com/mlengine/jobs/{}?project={}".format(
+            job_id, project_id
+        )
+    )
 
 
 def _stream_logs(job_id):
@@ -166,17 +202,26 @@ def _stream_logs(job_id):
         RuntimeError: if there are any errors from the streaming subprocess.
     """
     try:
-        print('Streaming job logs: ')
-        subprocess.run(
-            ["gcloud", "ai-platform", "jobs", "stream-logs", job_id])
+        print("Streaming job logs: ")
+        process = subprocess.Popen(
+            ["gcloud", "ai-platform", "jobs", "stream-logs", job_id],
+            stdout=subprocess.PIPE,
+        )
+        while True:
+            output = process.stdout.readline()
+            # Break out of the loop when poll returns an exit code.
+            if process.poll() is not None:
+                break
+            if output:
+                print(output.decode().replace("\x08", ""))
     except subprocess.SubprocessError as err:
         raise RuntimeError(
-            'There was an error streaming the job logs. {}'.format(
-                err._get_reason()))
+            "There was an error streaming the job logs. {}".format(err._get_reason())
+        )
 
 
 def _generate_job_id():
     """Returns a unique job id prefixed with 'tf_train'."""
     # CAIP job id can contains only numbers, letters and underscores.
-    unique_tag = str(uuid.uuid4()).replace('-', '_')
-    return 'tf_train_{}'.format(unique_tag)
+    unique_tag = str(uuid.uuid4()).replace("-", "_")
+    return "tf_cloud_train_{}".format(unique_tag)
