@@ -24,29 +24,35 @@ from . import containerize
 from . import deploy
 from . import gcp
 from . import machine_config
-from . import package
 from . import preprocess
 from . import validate
 
 
-def run(entry_point=None,
-        requirements_txt=None,
-        distribution_strategy='auto',
-        docker_base_image=None,
-        chief_config='auto',
-        worker_config='auto',
-        worker_count=0,
-        entry_point_args=None,
-        stream_logs=False):
+def run(
+    entry_point=None,
+    requirements_txt=None,
+    distribution_strategy="auto",
+    docker_base_image=None,
+    chief_config="auto",
+    worker_config="auto",
+    worker_count=0,
+    entry_point_args=None,
+    stream_logs=False,
+    docker_image_bucket_name=None,
+    **kwargs
+):
     """Runs your Tensorflow code in Google Cloud Platform.
 
     Args:
-        entry_point: Optional string. Python file path to the file that
-            contains the TensorFlow code.
+        entry_point: Optional string. File path to the python file or iPython
+            notebook that contains the TensorFlow code.
             Note: This path must be in the current working directory tree.
-            Example: 'train.py', 'training/mnist.py'
-            If `entry_point` is not provided, then the current python script is
-            assumed to be the `entry_point`.
+            Example: 'train.py', 'training/mnist.py', 'mnist.ipynb'
+            If `entry_point` is not provided, then
+            - If you are in an iPython notebook environment, then the
+                current notebook is taken as the `entry_point`.
+            - Otherwise, the current python script is taken as the
+                `entry_point`.
         requirements_txt: Optional string. File path to requirements.txt file
             containing aditional pip dependencies if any. ie. a file with a
             list of pip dependency package names.
@@ -58,7 +64,9 @@ def run(entry_point=None,
             you have provided using the `chief_config`, `worker_config` and
             `worker_count` params.
             - If the number of workers > 0, we will use
-                `tf.distribute.experimental.MultiWorkerMirroredStrategy`.
+                `tf.distribute.experimental.MultiWorkerMirroredStrategy` or
+                `tf.distribute.experimental.TPUStrategy` based on the
+                accelerator type.
             - If number of GPUs > 0, we will use
                 `tf.distribute.MirroredStrategy`
             - Otherwise, we will use `tf.distribute.OneDeviceStrategy`
@@ -81,94 +89,132 @@ def run(entry_point=None,
             Defaults to 'auto'. 'auto' maps to a standard gpu config such as
             `COMMON_MACHINE_CONFIGS.P100_1X` (8 cpu cores, 30GB memory,
             1 Nvidia Tesla P100).
+            For TPU strategy, `chief_config` refers to the config of the host
+            that controls the TPU workers.
         worker_config: Optional `MachineConfig` that represents the
             configuration for the general workers in a distribution cluster.
             Defaults to 'auto'. 'auto' maps to a standard gpu config such as
             `COMMON_MACHINE_CONFIGS.P100_1X` (8 cpu cores, 30GB memory,
             1 Nvidia Tesla P100).
+            For TPU strategy, `worker_config` should be a TPU config with 
+            8 TPU cores (eg. `COMMON_MACHINE_CONFIGS.TPU`).
         worker_count: Optional integer that represents the number of general
             workers in a distribution cluster. Defaults to 0. This count does
             not include the chief worker.
+            For TPU strategy, `worker_count` should be set to 1.
         entry_point_args: Optional list of strings. Defaults to None.
             Command line arguments to pass to the `entry_point` program.
         stream_logs: Boolean flag which when enabled streams logs back from
             the cloud job.
+        docker_image_bucket_name: Optional string that specifies the docker
+            image cloud storage bucket name. If this parameter is set, then we
+            will be using Google Cloud Storage and Google Cloud Build for
+            docker containerization (https://cloud.google.com/cloud-build/).
+            If this parameter is not set, then we will use local docker daemon
+            process for containerization.
+            Cloud build request is queued and we make a maximum of 10 requests
+            with a delay of 30 secs in between these requests to inspect the
+            status of the build.
+            Note: When you are using this API from within an iPython notebook,
+            we will default to using Google Cloud Build,
+            so `docker_image_bucket_name` must be specified for this use case.
+        **kwargs: Additional keyword arguments.
     """
     # If code is triggered in a cloud environment, do nothing.
-    if os.environ.get('TF_KERAS_RUNNING_REMOTELY'):
+    if os.environ.get("TF_KERAS_RUNNING_REMOTELY"):
         return
 
+    if kwargs:
+        # We are using kwargs for forward compatibility in the cloud. For eg.,
+        # if a new param is added to `run` API, this will not exist in the
+        # latest tensorflow-cloud package installed in the cloud docker envs.
+        # So, if `run` is used inside a python script or notebook, this python
+        # code will fail to run in the cloud even before we can check
+        # `TF_KERAS_RUNNING_REMOTELY` env var because of an additional unknown
+        # param.
+        raise TypeError("Unknown keyword arguments: %s" % (kwargs.keys(),))
+
     # Get defaults values for input param
 
     # If `entry_point` is not provided it means that the `run` API call
-    # is embedded in the script that contains the Keras module. For this
-    # script to run successfully in the cloud env, `tensorflow-cloud` pip
+    # is embedded in the script/notebook that contains the Keras module.
+    # For this to run successfully in the cloud env, `tensorflow-cloud` pip
     # package is required to be installed in addition to the user provided
     # packages.
-    install_tf_cloud = False
-    if entry_point is None:
-        entry_point = sys.argv[0]
-        install_tf_cloud = True
-    if chief_config == 'auto':
-        chief_config = machine_config.COMMON_MACHINE_CONFIGS['P100_1X']
-    if worker_config == 'auto':
-        worker_config = machine_config.COMMON_MACHINE_CONFIGS['P100_1X']
+    if chief_config == "auto":
+        chief_config = machine_config.COMMON_MACHINE_CONFIGS["P100_1X"]
+    if worker_config == "auto":
+        worker_config = machine_config.COMMON_MACHINE_CONFIGS["P100_1X"]
     region = gcp.get_region()
     # Working directory in the docker container filesystem.
-    destination_dir = '/app/'
+    destination_dir = "/app/"
     if not isinstance(worker_count, int):
         worker_count = int(worker_count)
     # Default location to which the docker image that is created is pushed.
-    docker_registry = 'gcr.io/{}'.format(gcp.get_project_name())
+    docker_registry = "gcr.io/{}".format(gcp.get_project_name())
+    called_from_notebook = _called_from_notebook()
 
     # Run validations.
     validate.validate(
-        entry_point, requirements_txt, distribution_strategy,
-        chief_config, worker_config, worker_count, region,
-        entry_point_args, stream_logs)
+        entry_point,
+        requirements_txt,
+        distribution_strategy,
+        chief_config,
+        worker_config,
+        worker_count,
+        region,
+        entry_point_args,
+        stream_logs,
+        docker_image_bucket_name,
+        called_from_notebook,
+    )
 
     # Make the `entry_point` cloud and distribution ready.
-    # A temporary script called `wrapped_entry_point` is created.
+    # A temporary script called `preprocessed_entry_point` is created.
     # This contains the `entry_point` wrapped in distribution strategy.
-    wrapped_entry_point = None
-    if (distribution_strategy == 'auto' and
-        chief_config.accelerator_type !=
-            machine_config.AcceleratorType.NO_ACCELERATOR):
-        wrapped_entry_point = preprocess.get_wrapped_entry_point(
-            entry_point, chief_config, worker_count)
+    preprocessed_entry_point = None
+    if (
+        distribution_strategy == "auto"
+        or entry_point.endswith("ipynb")
+        or entry_point is None
+    ):
+        preprocessed_entry_point = preprocess.get_preprocessed_entry_point(
+            entry_point,
+            chief_config,
+            worker_config,
+            worker_count,
+            distribution_strategy,
+            called_from_notebook=called_from_notebook,
+        )
 
-    # Create docker file.
-    docker_entry_point = wrapped_entry_point or entry_point
-    docker_file = containerize.create_docker_file(
-        docker_entry_point,
-        chief_config,
-        requirements_txt=requirements_txt,
-        destination_dir=destination_dir,
-        docker_base_image=docker_base_image,
-        install_tf_cloud=install_tf_cloud)
-
-    # Get all the files, that we need to package, mapped to the destination
-    # location. This will include the wrapped_entry_point, requirements_txt,
-    # dockerfile and the entry_point directory tree.
-    file_map = containerize.get_file_path_map(
+    # Create docker file, geenrate a tarball, build and push docker
+    # image using the tarball.
+    cb_args = (
         entry_point,
-        docker_file,
-        wrapped_entry_point=wrapped_entry_point,
-        requirements_txt=requirements_txt,
-        destination_dir=destination_dir)
-
-    # Create a tarball with the files.
-    tar_file_path = package.get_tar_file_path(file_map)
-
-    # Build and push docker image.
-    docker_img_uri = containerize.get_docker_image(
-        docker_registry, tar_file_path)
+        preprocessed_entry_point,
+        chief_config,
+        worker_config,
+        docker_registry,
+        gcp.get_project_name(),
+    )
+    cb_kwargs = {
+        "requirements_txt": requirements_txt,
+        "destination_dir": destination_dir,
+        "docker_base_image": docker_base_image,
+        "docker_image_bucket_name": docker_image_bucket_name,
+        "called_from_notebook": called_from_notebook,
+    }
+    if docker_image_bucket_name is None:
+        container_builder = containerize.LocalContainerBuilder(*cb_args, **cb_kwargs)
+    else:
+        container_builder = containerize.CloudContainerBuilder(*cb_args, **cb_kwargs)
+    docker_img_uri = container_builder.get_docker_image()
 
     # Delete all the temporary files we created.
-    if wrapped_entry_point is not None:
-        os.remove(wrapped_entry_point)
-    os.remove(docker_file)
-    os.remove(tar_file_path)
+    if preprocessed_entry_point is not None:
+        os.remove(preprocessed_entry_point)
+    for f in container_builder.get_generated_files():
+        os.remove(f)
 
     # Deploy docker image on the cloud.
     job_name = deploy.deploy_job(
@@ -178,8 +224,22 @@ def run(entry_point=None,
         worker_count,
         worker_config,
         entry_point_args,
-        stream_logs)
+        stream_logs,
+    )
 
     # Call `exit` to prevent training the Keras model in the local env.
     # To stop execution after encountering a `run` API call in local env.
-    sys.exit(0)
+    if not called_from_notebook:
+        sys.exit(0)
+
+
+def _called_from_notebook():
+    """Detects if we are currently executing in a notebook environment."""
+    try:
+        shell = get_ipython().__class__.__name__
+        if "Shell" in shell:
+            return True
+        else:
+            return False
+    except NameError:
+        return False
